import asyncio
import logging
from contextlib import asynccontextmanager
from importlib.resources import files

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from formica.settings import app_config
from formica.web.api_routes.main import api_router
from formica.web.web_routes import web_router
from starlette.staticfiles import StaticFiles

@asynccontextmanager
async def lifespan(app_: FastAPI):
    print("Starting up")
    yield
    print("Shutting down")


app = FastAPI(lifespan=lifespan)
logger = logging.getLogger(__name__)

origins = [
    "*",
]
_assets_dir = files("formica.web.dist").joinpath("assets")
app.mount(
    "/assets", StaticFiles(directory=str(_assets_dir)), name="assets"
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")
app.include_router(web_router, prefix="")


def run_webserver():
    import uvicorn
    try:
        uvicorn.run(
            "formica.web.main:app",
            host=app_config.get("webserver", "HOST"),
            port=app_config.getint("webserver", "PORT"),
            reload=False,
        )
    except KeyboardInterrupt:
        logger.info("Uvicorn server interrupted. Exiting gracefully.")
    except asyncio.CancelledError:
        print("Cancelled during shutdown")


if __name__ == "__main__":
    run_webserver()



import logging
import time
from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Query
from formica.models.models import FlowRunModel
from formica.models.models import FlowVersionModel
from formica.models.models import User
from formica.models.request_models import CreateFlowRun
from formica.models.request_models import FlowRunFilters
from formica.models.request_models import Log
from formica.models.request_models import ResponseDeviceRun
from formica.models.request_models import ResponseFlowRun
from formica.settings import FORMICA_HOME
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN
from starlette.status import HTTP_404_NOT_FOUND

SessionDep = Annotated[AsyncSession, Depends(get_session)]
flow_run_router = APIRouter()
logger = logging.getLogger(__name__)


@flow_run_router.post(
    path="", description="Create a FlowRun", status_code=HTTP_201_CREATED
)
async def add_flow_run(
    create_flow_run: CreateFlowRun,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> FlowRunModel:
    # new_flow_run = FlowRunModel.model_validate(new_flow_run)
    new_flow_run = FlowRunModel(
        flow_run_id=f"manual__{int(time.time())}",
        flow_id=create_flow_run.flow_id,
        version=create_flow_run.version,
        device_set_id=create_flow_run.device_set_id,
        description=create_flow_run.description,
        args=create_flow_run.args,
        run_type=create_flow_run.run_type,
        owner_id=user.id,
    )
    new_flow_run.flow_version = await FlowVersionModel.get_by_key(
        new_flow_run.flow_id, new_flow_run.version
    )

    flow_version = await FlowVersionModel.get_by_key(
        new_flow_run.flow_id, new_flow_run.version
    )

    if flow_version is None:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail="Flow version not found"
        )

    if await FlowRunModel.flow_run_existed(
        new_flow_run.flow_id, new_flow_run.version, new_flow_run.flow_run_id
    ):
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Flow run existed")

    if flow_version.flow.group_id not in (await user.get_group_ids()):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="This flow version is not in your group, you cannot create a flow run for it",
        )
    session.expunge(user)
    session.add(new_flow_run)
    await session.commit()
    await session.refresh(new_flow_run)

    return new_flow_run


@flow_run_router.get(path="", description="Get flow runs", status_code=HTTP_200_OK)
async def get_flow_runs(
    session: SessionDep,
    filters: Annotated[FlowRunFilters, Query()],
    user: User = Depends(current_active_user),
) -> list[ResponseFlowRun]:
    db_flow_runs: list[FlowRunModel] = await FlowRunModel.filter(user, filters, session)
    response = []

    for db_flow_run in db_flow_runs:
        # Create the ResponseDeviceRun objects
        response_device_runs = []
        for device_run in db_flow_run.device_runs:
            # Get the logs
            log_file = (
                FORMICA_HOME
                / "logs"
                / f"flow_id={device_run.flow_id}"
                / f"version={device_run.version}"
                / f"flow_run={device_run.flow_run_id}"
                / f"{device_run.device_id}.log"
            )

            logs = []
            try:
                with open(log_file, "r") as f:
                    logs = f.readlines()
            except FileNotFoundError as e:
                logger.warning(f"Can't find log file at {log_file}, exception: {e}")

            response_device_run = ResponseDeviceRun(
                id=device_run.device_run_db_id,
                flow_id=device_run.flow_id,
                version=device_run.version,
                flow_run_id=device_run.flow_run_id,
                device_id=device_run.device_id,
                state=device_run.state,
                logical_start_time=device_run.logical_start_time,
                actual_start_time=device_run.actual_start_time,
                end_time=device_run.end_time,
                created_at=device_run.created_at,
                logs=[
                    Log(timestamp="", level="info", message=log) for log in logs
                ],  # Add logs to the response
            )
            response_device_runs.append(response_device_run)

        response_flow_run = ResponseFlowRun(
            flow_run_id=db_flow_run.flow_run_id,
            flow_id=db_flow_run.flow_id,
            version=db_flow_run.version,
            description=db_flow_run.description,
            device_set_id=db_flow_run.device_set_id,
            args=db_flow_run.args,
            run_type=db_flow_run.run_type,
            start_time=db_flow_run.start_time,
            end_time=db_flow_run.end_time,
            state=db_flow_run.state,
            created_at=db_flow_run.created_at,
            device_runs=response_device_runs,
        )

        response.append(response_flow_run)

    return response



import uuid
from datetime import datetime
from typing import Optional

from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseUserTableUUID
from fastapi_users.db import SQLAlchemyUserDatabase
from formica.models.constant import ConnectionType
from formica.models.constant import DeviceRunState
from formica.models.constant import DeviceType
from formica.models.constant import FlowRunState
from formica.models.constant import FlowRunType
from formica.models.constant import INITIAL_FLOW_STATE
from formica.models.constant import TaskState
from formica.models.constant import TaskTriggerRule
from formica.models.constant import UserRole
from formica.models.request_models import CredentialFilters
from formica.models.request_models import DeviceFilters
from formica.models.request_models import DeviceSetFilters
from formica.models.request_models import FlowFilters
from formica.models.request_models import FlowRunFilters
from formica.models.request_models import FlowVersionFilters
from formica.models.request_models import GroupFilters
from formica.utils.session import get_session
from formica.utils.session import NEW_SESSION
from formica.utils.session import provide_session
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy import ForeignKeyConstraint
from sqlalchemy import func
from sqlalchemy import JSON
from sqlalchemy import Select
from sqlalchemy import String
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import declarative_base
from sqlmodel import desc
from sqlmodel import Field
from sqlmodel import Relationship
from sqlmodel import select
from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession

Base = declarative_base(metadata=SQLModel.metadata)
UUID_ID = uuid.UUID


# Junction table for many-to-many relationship
class UserGroupLink(SQLModel, table=True):
    __tablename__ = "user_group_link"
    email: str = Field(foreign_key="user.email", primary_key=True, ondelete="CASCADE")
    group_id: str = Field(
        foreign_key="user_group.group_id", primary_key=True, ondelete="CASCADE"
    )


class User(SQLAlchemyBaseUserTableUUID, Base):
    name = Column(String)
    role = Column(Enum(UserRole))
    groups: list["GroupModel"] = Relationship(
        back_populates="members", link_model=UserGroupLink
    )
    created_at = Column(DateTime(timezone=False), server_default=func.now())

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = Depends(get_session)):
        stmt: Select = select(User)
        return (await session.exec(stmt)).unique().all()

    @provide_session
    async def get_group_ids(self, session: AsyncSession = Depends(get_session)):
        stmt: Select = select(UserGroupLink.group_id).where(
            UserGroupLink.email == self.email
        )
        return (await session.exec(stmt)).all()

    @classmethod
    @provide_session
    async def get_user_by_email(cls, email: str, session: AsyncSession = NEW_SESSION):
        stmt: Select = select(User).where(User.email == email)
        return (await session.exec(stmt)).first()

    @classmethod
    @provide_session
    async def get_user_by_ids(
        cls, user_ids: list[str], session: AsyncSession = NEW_SESSION
    ):
        stmt: Select = select(User).where(User.id.in_(user_ids))
        return (await session.exec(stmt)).all()

    @classmethod
    @provide_session
    async def get_user_by_emails(
        cls, emails: list[str], session: AsyncSession = NEW_SESSION
    ):
        stmt: Select = select(User).where(User.email.in_(emails))
        return (await session.exec(stmt)).all()


async def get_user_db(session: AsyncSession = Depends(get_session)):
    yield SQLAlchemyUserDatabase(session, User)


class GroupModel(SQLModel, table=True):
    __tablename__ = "user_group"
    group_id: str = Field(primary_key=True)
    description: str = Field(default="")

    devices: list["DeviceModel"] = Relationship(
        back_populates="group",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )
    device_sets: list["DeviceSetModel"] = Relationship(
        back_populates="group",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )
    flows: list["FlowModel"] = Relationship(back_populates="group")
    members: list[User] = Relationship(
        link_model=UserGroupLink, sa_relationship_kwargs={"lazy": "joined"}
    )

    @classmethod
    @provide_session
    async def filter(
        cls, user: User, filters: GroupFilters, session: AsyncSession = NEW_SESSION
    ) -> list["GroupModel"]:
        query: Select = select(GroupModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(DeviceModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        group_ids = await user.get_group_ids(session)

        if not user.is_superuser:
            query = query.where(GroupModel.group_id.in_(group_ids))

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = Depends(get_session)):
        stmt: Select = select(GroupModel)
        return (await session.exec(stmt)).unique().all()

    @provide_session
    async def add_members(
        self, members: list[User], session: AsyncSession = NEW_SESSION
    ):
        for member in members:
            stmt: Select = select(UserGroupLink).where(
                UserGroupLink.id == member.id,
                UserGroupLink.group_id == self.group_id,
            )
            if not (await session.exec(stmt)).first():
                user_group_link = UserGroupLink(id=member.id, group_id=self.group_id)
                session.add(user_group_link)

    @classmethod
    @provide_session
    async def get_members(cls, group, session: AsyncSession = NEW_SESSION):
        stmt: Select = (
            select(User)
            .join(UserGroupLink)
            .where(UserGroupLink.group_id == group.group_id)
        )
        return (await session.exec(stmt)).all()

    @classmethod
    @provide_session
    async def group_existed(
        cls, group_id: str, session: AsyncSession = NEW_SESSION
    ) -> bool:
        return (await GroupModel.get_by_key(group_id, session)) is not None

    @classmethod
    @provide_session
    async def get_by_key(
        cls, group_id: str, session: AsyncSession = NEW_SESSION
    ) -> Optional["GroupModel"]:
        return await session.get(GroupModel, group_id)


class FlowModel(SQLModel, table=True):
    __tablename__ = "flow"
    flow_id: str = Field(default="", primary_key=True)
    owner_id: UUID_ID = Field(foreign_key="user.id")
    group_id: str = Field(foreign_key="user_group.group_id")
    saved_state: dict = Field(sa_column=Column(JSON), default=INITIAL_FLOW_STATE)
    description: str = Field(default="")
    created_at: datetime = Field(default_factory=datetime.now)

    group: GroupModel = Relationship(
        back_populates="flows", sa_relationship_kwargs={"lazy": "selectin"}
    )

    @classmethod
    @provide_session
    async def filter(
        cls, user: User, filters: FlowFilters, session: AsyncSession = NEW_SESSION
    ):
        query: Select = select(FlowModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(DeviceModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                FlowModel.group_id.in_(await user.get_group_ids(session))
            )

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def flow_existed(
        cls, flow_id: str, session: AsyncSession = NEW_SESSION
    ) -> bool:
        return (await FlowModel.get_by_key(flow_id, session)) is not None

    @classmethod
    @provide_session
    async def get_by_key(cls, flow_id: str, session: AsyncSession = NEW_SESSION):
        return await session.get(FlowModel, flow_id)

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = NEW_SESSION):
        stmt: Select = select(FlowModel)
        return (await session.exec(stmt)).unique().all()


class FlowVersionModel(SQLModel, table=True):
    __tablename__ = "flow_version"
    flow_id: str = Field(primary_key=True, foreign_key="flow.flow_id")
    version: str = Field(primary_key=True)
    parameters: list[str] = Field(sa_column=Column(JSON))
    structure: dict = Field(sa_column=Column(JSON))
    description: str = Field(default="")
    created_at: datetime = Field(default_factory=datetime.now)

    flow: FlowModel = Relationship(sa_relationship_kwargs={"lazy": "selectin"})

    @classmethod
    @provide_session
    async def filter(
        cls,
        user: User,
        filters: FlowVersionFilters,
        session: AsyncSession = NEW_SESSION,
    ):
        query: Select = select(FlowVersionModel).join(FlowModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(FlowVersionModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                FlowModel.group_id.in_(await user.get_group_ids(session))
            )

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def flow_version_existed(
        cls, flow_id: str, version: str, session: AsyncSession = NEW_SESSION
    ) -> bool:
        return (
            await FlowVersionModel.get_by_key(flow_id, version, session)
        ) is not None

    @classmethod
    @provide_session
    async def get_by_key(
        cls, flow_id: str, version: str, session: AsyncSession = NEW_SESSION
    ):
        return await session.get(FlowVersionModel, (flow_id, version))


class FlowRunModel(SQLModel, table=True):
    __table_args__ = (
        ForeignKeyConstraint(
            ["flow_id", "version"],
            ["flow_version.flow_id", "flow_version.version"],
        ),
    )
    __tablename__ = "flow_run"
    flow_id: str = Field(primary_key=True)
    version: str = Field(primary_key=True)
    flow_run_id: str = Field(primary_key=True)
    owner_id: UUID_ID = Field(foreign_key="user.id")
    description: str = Field(default="")
    device_set_id: str = Field(foreign_key="device_set.device_set_id")
    args: Optional[dict[str, str]] = Field(sa_column=Column(JSON))
    run_type: FlowRunType = Field(sa_column=Column(Enum(FlowRunType)))
    start_time: datetime = Field(default_factory=datetime.now)
    end_time: Optional[datetime] = Field()
    state: FlowRunState = Field(default=FlowRunState.SUBMITTED)
    created_at: datetime = Field(default_factory=datetime.now)

    device_runs: list["DeviceRunModel"] = Relationship(
        back_populates="flow_run",
        sa_relationship_kwargs={"lazy": "selectin"},
    )
    flow_version: FlowVersionModel = Relationship(
        sa_relationship_kwargs={"lazy": "selectin"}
    )

    @classmethod
    @provide_session
    async def filter(
        cls,
        user: User,
        filters: FlowRunFilters,
        session: AsyncSession = NEW_SESSION,
    ):
        query: Select = select(FlowRunModel).join(FlowVersionModel).join(FlowModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(FlowRunModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                FlowModel.group_id.in_(await user.get_group_ids(session))
            ).where(FlowRunModel.owner_id == user.id)

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def flow_run_existed(
        cls,
        flow_id: str,
        version: str,
        flow_run_id: str,
        session: AsyncSession = NEW_SESSION,
    ) -> bool:
        return (
            await FlowRunModel.get_by_key(flow_id, version, flow_run_id, session)
        ) is not None

    @classmethod
    @provide_session
    async def get_by_key(
        cls,
        flow_id: str,
        version: str,
        flow_run_id: str,
        session: AsyncSession = NEW_SESSION,
    ):
        return await session.get(FlowRunModel, (flow_id, version, flow_run_id))

    @classmethod
    @provide_session
    async def get_flowrun_to_schedule(cls, session: AsyncSession = NEW_SESSION):
        """Lấy các FlowRun có trạng thái SUBMITTED hoặc RUNNING để schedule"""
        stmt: Select = select(FlowRunModel).where(
            (FlowRunModel.state == FlowRunState.SUBMITTED)
            | (FlowRunModel.state == FlowRunState.RUNNING)
        )
        return (await session.exec(stmt)).all()

    def set_state(self, new_state: FlowRunState):
        if new_state == FlowRunState.SUBMITTED:
            raise ValueError("Can't set a FlowRun's state to SUBMITTED")
        elif new_state == FlowRunState.RUNNING:
            self.start_time = datetime.now()
            self.state = new_state
        elif new_state == FlowRunState.FINISHED:
            self.end_time = datetime.now()
            self.state = new_state
        else:
            raise ValueError(f"Invalid FlowRunState: {new_state}")


# Junction table for many-to-many relationship
class DeviceSetLink(SQLModel, table=True):
    __tablename__ = "device_set_link"

    device_id: str = Field(foreign_key="device.device_id", primary_key=True)
    device_set_id: str = Field(foreign_key="device_set.device_set_id", primary_key=True)


class DeviceModel(SQLModel, table=True):
    __tablename__ = "device"
    group_id: str = Field(primary_key=True, foreign_key="user_group.group_id")
    device_id: str = Field(primary_key=True)
    device_type: DeviceType = Field()
    ip: str = Field()

    group: GroupModel = Relationship(back_populates="devices")
    device_sets: list["DeviceSetModel"] = Relationship(
        back_populates="devices", link_model=DeviceSetLink
    )
    credentials: list["CredentialModel"] = Relationship(
        back_populates="device",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )

    @classmethod
    @provide_session
    async def filter(
        cls, user: User, filters: DeviceFilters, session: AsyncSession = NEW_SESSION
    ):
        query: Select = select(DeviceModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(DeviceModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                DeviceModel.group_id.in_(await user.get_group_ids(session))
            )

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def get_devices_by_id(
        cls, device_ids: list[str], session: AsyncSession = NEW_SESSION
    ):
        stmt: Select = select(DeviceModel).where(DeviceModel.device_id.in_(device_ids))
        return (await session.exec(stmt)).all()

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = NEW_SESSION):
        stmt: Select = select(DeviceModel)
        return (await session.exec(stmt)).unique().all()

    @classmethod
    @provide_session
    async def device_existed(
        cls, group_id: str, dev_id: str, session: AsyncSession = NEW_SESSION
    ) -> bool:
        return (await DeviceModel.get_by_key(group_id, dev_id, session)) is not None

    @classmethod
    @provide_session
    async def get_by_key(
        cls, group_id: str, dev_id: str, session: AsyncSession = NEW_SESSION
    ):
        return await session.get(DeviceModel, (group_id, dev_id))


class DeviceSetModel(SQLModel, table=True):
    __tablename__ = "device_set"
    device_set_id: str = Field(primary_key=True)
    description: str = Field(default="")
    group_id: str = Field(foreign_key="user_group.group_id")

    devices: list[DeviceModel] = Relationship(
        back_populates="device_sets",
        link_model=DeviceSetLink,
        sa_relationship_kwargs={"lazy": "selectin"},
    )
    group: GroupModel = Relationship(back_populates="device_sets")

    @classmethod
    @provide_session
    async def filter(
        cls, user: User, filters: DeviceSetFilters, session: AsyncSession = NEW_SESSION
    ):
        query: Select = select(DeviceSetModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(DeviceSetModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                DeviceSetModel.group_id.in_(await user.get_group_ids(session))
            )

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def get_by_key(cls, device_set_id: str, session: AsyncSession = NEW_SESSION):
        return await session.get(DeviceSetModel, device_set_id)

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = NEW_SESSION):
        stmt: Select = select(DeviceSetModel)
        return (await session.exec(stmt)).unique().all()


class CredentialModel(SQLModel, table=True):
    __tablename__ = "credential"

    __table_args__ = (
        UniqueConstraint("group_id", "device_id", "connection_type", "priority"),
    )

    id: Optional[int] = Field(primary_key=True)
    device_id: str = Field(foreign_key="device.device_id")
    group_id: str = Field(foreign_key="user_group.group_id")
    connection_type: ConnectionType = Field()
    priority: int = Field()
    description: str = Field()
    username: str = Field()
    password: Optional[str] = Field()
    port: int = Field()
    extra: dict = Field(sa_column=Column(JSON))

    device: DeviceModel = Relationship(
        back_populates="credentials", sa_relationship_kwargs={"lazy": "selectin"}
    )

    @classmethod
    @provide_session
    async def filter(
        cls, user: User, filters: CredentialFilters, session: AsyncSession = NEW_SESSION
    ):
        query: Select = select(CredentialModel).join(DeviceModel)

        for field_name, value in filters.model_dump(exclude_none=True).items():
            column = getattr(DeviceModel, field_name)
            query = query.where(column == value)
            print(f"{column} = {value}")

        if not user.is_superuser:
            query = query.where(
                DeviceModel.group_id.in_(await user.get_group_ids(session))
            )

        return (await session.exec(query)).unique().all()

    @classmethod
    @provide_session
    async def get_all(cls, session: AsyncSession = Depends(get_session)):
        stmt: Select = select(CredentialModel)
        return (await session.exec(stmt)).unique().all()

    @classmethod
    @provide_session
    async def credential_existed(
        cls,
        group_id: str,
        device_id: str,
        connection_type: ConnectionType,
        priority: int,
        session: AsyncSession = Depends(get_session),
    ):
        stmt: Select = select(CredentialModel).where(
            CredentialModel.group_id == group_id,
            CredentialModel.device_id == device_id,
            CredentialModel.connection_type == connection_type,
            CredentialModel.priority == priority,
        )
        return (await session.exec(stmt)).first() is not None

    @classmethod
    @provide_session
    async def get_by_key(cls, credential_id: str, session: AsyncSession = NEW_SESSION):
        return await session.get(CredentialModel, credential_id)


class DeviceRunModel(SQLModel, table=True):
    __table_args__ = (
        ForeignKeyConstraint(
            ["flow_id", "version", "flow_run_id"],
            ["flow_run.flow_id", "flow_run.version", "flow_run.flow_run_id"],
        ),
    )
    __tablename__ = "device_run"
    device_run_db_id: Optional[int] = Field(primary_key=True)
    flow_id: str = Field()
    version: str = Field()
    flow_run_id: str = Field()
    device_id: str = Field(foreign_key="device.device_id")
    state: DeviceRunState = Field(default=DeviceRunState.QUEUED)
    logical_start_time: datetime = Field()
    actual_start_time: Optional[datetime] = Field()
    end_time: Optional[datetime] = Field()
    created_at: datetime = Field(default_factory=datetime.now)

    device: DeviceModel = Relationship(sa_relationship_kwargs={"lazy": "selectin"})
    flow_run: FlowRunModel = Relationship(
        back_populates="device_runs", sa_relationship_kwargs={"lazy": "selectin"}
    )
    tasks: list["TaskModel"] = Relationship(
        back_populates="device_run", sa_relationship_kwargs={"lazy": "selectin"}
    )

    @classmethod
    @provide_session
    async def get_latest_device_run_of_flowrun(
        cls, flowrun: FlowRunModel, session: AsyncSession = NEW_SESSION
    ):
        stmt: Select = (
            select(DeviceRunModel)
            .where(
                (DeviceRunModel.flow_id == flowrun.flow_id)
                & (DeviceRunModel.version == flowrun.version)
                & (DeviceRunModel.flow_run_id == flowrun.flow_run_id)
            )
            .order_by(desc(DeviceRunModel.actual_start_time))
        )
        return (await session.exec(stmt)).first()

    def set_state(self, new_state: DeviceRunState):
        if new_state == DeviceRunState.QUEUED:
            raise ValueError("Can't set a DeviceRun's state to QUEUED")
        elif new_state == DeviceRunState.RUNNING:
            self.state = new_state
            self.actual_start_time = datetime.now()
        elif new_state in (DeviceRunState.SUCCESS, DeviceRunState.FAILED):
            self.state = new_state
            self.end_time = datetime.now()
        else:
            raise ValueError(f"Invalid DeviceRunState: {new_state}")


class TaskModel(SQLModel, table=True):
    __tablename__ = "task"
    device_run_db_id: str = Field(
        primary_key=True, foreign_key="device_run.device_run_db_id"
    )
    node_id: str = Field(primary_key=True)
    state: TaskState = Field(default=TaskState.WAIT_FOR_EXECUTING)
    trigger_rule: TaskTriggerRule = Field(default=TaskTriggerRule.SUCCESS)
    created_at: datetime = Field(default_factory=datetime.now)
    start_time: Optional[datetime] = Field()
    end_time: Optional[datetime] = Field()

    device_run: DeviceRunModel = Relationship(back_populates="tasks")

    @classmethod
    @provide_session
    async def get_tasks_by_flow_and_run_id(
        cls, flow_id: str, run_id: str, session: AsyncSession = NEW_SESSION
    ):
        stmt: Select = select(TaskModel).where(
            TaskModel.flow_id == flow_id and TaskModel.run_id == run_id
        )
        return (await session.exec(stmt)).all()

    def set_state(self, new_state: TaskState):
        if new_state == TaskState.WAIT_FOR_EXECUTING:
            raise ValueError("Can't set a Task's state to WAIT_FOR_EXECUTING")
        elif new_state == TaskState.RUNNING:
            self.state = new_state
            self.start_time = datetime.now()
        elif new_state in (TaskState.SUCCESS, TaskState.FAILED):
            self.state = new_state
            self.end_time = datetime.now()
        elif new_state == TaskState.SKIPPED:
            self.state = new_state
            self.start_time = None
            self.end_time = None
        else:
            raise ValueError(f"Invalid TaskState: {new_state}")



import uuid
from typing import Optional

from fastapi import Depends
from fastapi import Request
from fastapi_users import BaseUserManager
from fastapi_users import FastAPIUsers
from fastapi_users import models
from fastapi_users import UUIDIDMixin
from fastapi_users.authentication import AuthenticationBackend
from fastapi_users.authentication import BearerTransport
from fastapi_users.authentication import JWTStrategy
from fastapi_users.db import SQLAlchemyUserDatabase
from formica.models.models import get_user_db
from formica.models.models import User
from formica.settings import app_config

SECRET = app_config.get("security", "SECRET_KEY")
TOKEN_LIFETIME = app_config.getint("security", "TOKEN_LIFETIME")


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    reset_password_token_secret = SECRET
    verification_token_secret = SECRET

    async def on_after_register(self, user: User, request: Optional[Request] = None):
        print(f"User {user.id} has registered.")

    async def on_after_forgot_password(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        print(f"User {user.id} has forgot their password. Reset token: {token}")

    async def on_after_request_verify(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        print(f"Verification requested for user {user.id}. Verification token: {token}")


async def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


bearer_transport = BearerTransport(tokenUrl="auth/jwt/login")


def get_jwt_strategy() -> JWTStrategy[models.UP, models.ID]:
    return JWTStrategy(secret=SECRET, lifetime_seconds=TOKEN_LIFETIME)


auth_backend = AuthenticationBackend(
    name="jwt",
    transport=bearer_transport,
    get_strategy=get_jwt_strategy,
)

fastapi_users = FastAPIUsers[User, uuid.UUID](get_user_manager, [auth_backend])

current_active_user = fastapi_users.current_user(active=True)



from __future__ import annotations

import logging
import warnings
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Type
from typing import TYPE_CHECKING

from formica.models.constant import ConnectionType
from formica.models.constant import TaskTriggerRule
from jinja2 import Template

warnings.filterwarnings("ignore", category=DeprecationWarning)


if TYPE_CHECKING:
    from formica.execution.context import Context

logger = logging.getLogger(__name__)


class NodeFactory:
    _node_classes: dict[str, Type[BaseNode]] = {}

    @classmethod
    def register_node(cls, node_type: str, node_class: Type[BaseNode]):
        """Register a new node type."""
        cls._node_classes[node_type] = node_class

    @classmethod
    def create_node(cls, node_data: dict[str, Any]) -> BaseNode:
        """
        Create a node from the given JSON data.

        Example:
        {
            "node_id": "cmd_2",
            "type": "command",
            "command": "rm error.log",
            "timeout": 10,
            "trigger_rule": "success",
        }
        :param node_data: the JSON data
        :return: The node object created from the JSON data
        """
        # Remove the "type" field from node_data because the node's constructor doesn't need it
        node_args = {
            "node_id": node_data["id"],
            "timeout": 10,
            "trigger_rule": TaskTriggerRule.SUCCESS,
        }

        node_args.update(node_data["data"]["config"])
        node_type = node_data["data"]["label"].lower()

        if node_type not in cls._node_classes:
            raise ValueError(f"Unknown node type: {node_type}")

        return cls._node_classes[node_type](**node_args)


class BaseNode(ABC):
    # TODO: Use __new__() ?
    def __init__(
        self,
        node_id: str,
        timeout: int,
        trigger_rule: TaskTriggerRule,
        flow=None,
    ):
        self.node_id = node_id
        self.flow = flow
        self.timeout = timeout
        self.trigger_rule = trigger_rule
        self.upstream_node_ids = []
        self.downstream_node_ids = []
        self.output: str = ""

    @abstractmethod
    async def execute(self, context: Context):
        pass

    def add_downstream(self, node: BaseNode):
        self.downstream_node_ids.append(node.node_id)
        node.upstream_node_ids.append(self.node_id)

    def __str__(self):
        return __name__


class StartNode(BaseNode):
    async def execute(self, context: Context):
        pass


class ConnectNode(BaseNode):
    def __init__(self, connection_type: str, **kwargs):
        self.connection_type = ConnectionType(connection_type)
        super().__init__(**kwargs)

    async def execute(self, context: Context):
        """Connect to the device using the context given"""
        try:
            await context.make_connection(self.connection_type)
        except Exception as e:
            logger.warning(
                f"Can't connect to {context.device.device_id}. Error: {str(e)}"
            )
            raise e


class CommandNode(BaseNode):
    def __init__(self, command: str = "", **kwargs):
        super().__init__(**kwargs)
        self.command = command

    async def execute(self, context: Context):
        # Compile jinja2 để lấy được command thực cần chạy
        template = Template(self.command)
        vars = context.get_variables()
        compiled_command = template.render(vars)
        logger.debug(f"Original command: {self.command}")
        logger.debug(vars)
        logger.debug(f"Compiled command: {compiled_command}")
        print(f"Compiled command: {compiled_command}")

        try:
            self.output = await context.run_command(
                compiled_command, expect_prompt=None
            )

            # Update context
            context.set_node_output(self.node_id, self.output)
            context.append_log(self.output)
        except Exception as e:
            raise e


class DecisionNode(BaseNode):
    def __init__(self, condition: str = "", true_branch="", false_branch="", **kwargs):
        super().__init__(**kwargs)
        self.condition = condition
        self.true_branch = true_branch
        self.false_branch = false_branch
        self.output: str = ""
        self.skip_task = ""

    async def execute(self, context: Context):
        template = Template(self.condition)
        vars = context.get_variables()
        compiled_condition = template.render(vars)
        logger.info(f"Compiled condition: {compiled_condition}")
        # logger.info(f"output is : {context.flow.vars['ssh_command_ls']['output']}")
        condition_is_true = eval(compiled_condition)
        if condition_is_true:
            self.output = self.true_branch
            self.skip_task = self.false_branch
        else:
            self.output = self.false_branch
            self.skip_task = self.true_branch

        logger.info(f"condition output is: {self.output}")
        logger.info(f"condition skip task is: {self.skip_task}")


class ExitNode(BaseNode):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def execute(self, context: Context):
        try:
            await context.discard_connection()
        except Exception as e:
            logger.warning(f"Error while disconnecting: {e}")
            raise e


NodeFactory.register_node("start", StartNode)
NodeFactory.register_node("connect", ConnectNode)
NodeFactory.register_node("command", CommandNode)
NodeFactory.register_node("disconnect", ExitNode)
NodeFactory.register_node("decision", DecisionNode)



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Path
from fastapi import Query
from formica.models.constant import UserRole
from formica.models.models import FlowModel
from formica.models.models import FlowRunModel
from formica.models.models import FlowVersionModel
from formica.models.models import GroupModel
from formica.models.models import User
from formica.models.request_models import FlowFilters
from formica.models.request_models import ResponseFlow
from formica.models.request_models import UpdateFlow
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_204_NO_CONTENT
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN
from starlette.status import HTTP_404_NOT_FOUND

SessionDep = Annotated[AsyncSession, Depends(get_session)]
flow_router = APIRouter()


@flow_router.post(
    path="",
    description="Create a new flow",
    status_code=HTTP_201_CREATED,
)
async def add_flow(
    new_flow: FlowModel, session: SessionDep, user: User = Depends(current_active_user)
) -> FlowModel:
    if await FlowModel.flow_existed(new_flow.flow_id, session):
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Flow existed")

    group_model = await GroupModel.get_by_key(new_flow.group_id, session)

    if group_model is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Group not existed")

    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create a flow",
        )

    # Check if user is in the group
    if user not in group_model.members:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"User {user.email} is not a member of group {new_flow.group_id}",
        )

    new_flow.owner_id = user.id
    session.add(new_flow)
    await session.commit()
    await session.refresh(new_flow)

    return new_flow


@flow_router.get(path="", description="Get flows", status_code=HTTP_200_OK)
async def get_flows(
    filters: Annotated[FlowFilters, Query()],
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[FlowModel]:
    return await FlowModel.filter(user, filters, session)


@flow_router.get(
    path="/{flow_id}", description="Lấy flow theo flow id", status_code=HTTP_200_OK
)
async def get_flow(
    flow_id: str, session: SessionDep, user: User = Depends(current_active_user)
) -> ResponseFlow:
    flow = await FlowModel.get_by_key(flow_id, session)

    if flow is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Flow not found")

    if flow.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to access this flow",
        )

    return ResponseFlow.model_validate(flow.model_dump())


@flow_router.get(
    path="/{flow_id}/versions/{version}",
    description="Get a flow version",
    status_code=HTTP_200_OK,
)
async def get_flow_version(
    flow_id: str,
    version,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> FlowVersionModel:
    flow_version = await FlowVersionModel.get_by_key(flow_id, version, session)
    if not flow_version:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Version not found")

    if flow_version.flow.group_id not in user.get_group_ids(session):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to access this flow",
        )

    return flow_version


@flow_router.get(
    path="/{flow_id}/versions/{version}/flow-runs/{flow_run_id}",
    description="Get a flow run",
    status_code=HTTP_200_OK,
)
async def get_flow_run(
    flow_id: str,
    version: str,
    flow_run_id: str,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> FlowRunModel:
    flow_run = await FlowRunModel.get_by_key(flow_id, version, flow_run_id, session)
    if not flow_run:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Flow run not found")

    if flow_run.flow_version.flow_id.group_id not in user.get_group_ids(session):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to access this flow",
        )

    return flow_run


@flow_router.patch(path="/{target_flow_id}", description="Update flow")
async def update_flow(
    target_flow_id: Annotated[str, Path(title="The target flow id to update")],
    update: UpdateFlow,
    session: SessionDep,
    user: User = Depends(current_active_user),
):
    flow = await FlowModel.get_by_key(target_flow_id, session)
    if not flow:
        raise HTTPException(status_code=404, detail="Flow not found")

    # Only the owner or superuser can update the flow
    if flow.owner_id != user.id and not user.is_superuser:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to update this flow",
        )

    update_data = update.model_dump(exclude_unset=True)

    for key, value in update_data.items():
        setattr(flow, key, value)

    session.add(flow)
    await session.commit()
    await session.refresh(flow)
    return flow


@flow_router.delete(
    path="/{flow_id}",
    description="Delete a flow",
    status_code=HTTP_204_NO_CONTENT,
)
async def remove_device(
    flow_id: str,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> None:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to delete a flow",
        )

    flow = await FlowModel.get_by_key(flow_id, session)

    if flow.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Your groups do not have this flow: {flow_id}",
        )

    if flow is not None:
        await session.delete(flow)
        await session.commit()



import uuid
from datetime import datetime
from typing import Optional

from fastapi_users import schemas
from formica.models.constant import DeviceRunState
from formica.models.constant import FlowRunState
from formica.models.constant import FlowRunType
from formica.models.constant import UserRole
from pydantic import BaseModel


class CreateDeviceSet(BaseModel):
    device_set_id: str
    description: str
    devices: list[str]
    group_id: str


class UpdateUser(BaseModel):
    name: Optional[str] = None
    role: Optional[UserRole] = None
    email: Optional[str] = None
    is_active: Optional[str] = None


class UpdateFlow(BaseModel):
    flow_id: Optional[str] = None
    description: Optional[str] = None
    saved_state: dict


class UpdateUsersOfGroup(BaseModel):
    users: list[str]


class UpdateDevicesOfDeviceSet(BaseModel):
    devices: list[str]


class CreateFlowRun(BaseModel):
    flow_id: str
    version: str
    description: str
    device_set_id: str
    run_type: FlowRunType
    args: Optional[dict[str, str]] = None


class ResponseFlow(BaseModel):
    flow_id: str
    description: str
    saved_state: dict
    group_id: str


class Log(BaseModel):
    timestamp: str
    level: str
    message: str


class ResponseDeviceRun(BaseModel):
    id: int
    flow_id: str
    version: str
    flow_run_id: str
    device_id: str
    state: DeviceRunState
    logical_start_time: datetime
    actual_start_time: Optional[datetime]
    end_time: Optional[datetime]
    created_at: datetime
    logs: list[Log]


class ResponseFlowRun(BaseModel):
    flow_run_id: str
    flow_id: str
    version: str
    description: str
    device_set_id: str
    args: Optional[dict[str, str]] = None
    run_type: FlowRunType
    start_time: datetime
    end_time: Optional[datetime]
    state: FlowRunState
    created_at: datetime
    device_runs: list[ResponseDeviceRun]


class ResponseDeviceSet(BaseModel):
    device_set_id: str
    description: str
    devices: list[str]
    group_id: str


class UpdateDeviceSet(BaseModel):
    device_set_id: str
    devices: list[str]


class UserRead(schemas.BaseUser[uuid.UUID]):
    pass


class UserCreate(schemas.BaseUserCreate):
    name: str


class UserUpdate(schemas.BaseUserUpdate):
    pass


class ResponseUser(BaseModel):
    id: uuid.UUID
    name: str
    email: str
    role: UserRole
    is_superuser: bool


class GroupResponse(BaseModel):
    group_id: str
    description: str
    members: list[uuid.UUID]  # list of user ids


class DeviceSetResponse(BaseModel):
    device_set_id: str
    description: str
    devices: list[str]


# ============================= FILTERS ================================
class DeviceFilters(BaseModel):
    group_id: Optional[str] = None


class CredentialFilters(BaseModel):
    group_id: Optional[str] = None


class DeviceSetFilters(BaseModel):
    group_id: Optional[str] = None


class FlowFilters(BaseModel):
    group_id: Optional[str] = None


class FlowVersionFilters(BaseModel):
    flow_id: Optional[str] = None
    group_id: Optional[str] = None


class FlowRunFilters(BaseModel):
    flow_id: Optional[str] = None
    flow_version_id: Optional[str] = None
    group_id: Optional[str] = None


class GroupFilters(BaseModel):
    group_id: Optional[str] = None
    group_name: Optional[str] = None
    group_description: Optional[str] = None



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Query
from formica.models.constant import DeviceType
from formica.models.constant import UserRole
from formica.models.models import DeviceModel
from formica.models.models import User
from formica.models.request_models import DeviceFilters
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_204_NO_CONTENT
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN

SessionDep = Annotated[AsyncSession, Depends(get_session)]
device_router = APIRouter()


@device_router.post(
    path="", description="Create a device", status_code=HTTP_201_CREATED
)
async def add_device(
    new_device: DeviceModel,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> DeviceModel:
    if not DeviceType.is_valid(new_device.device_type):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Device type is not valid: {new_device.device_type}",
        )

    if await DeviceModel.device_existed(new_device.device_id, session):
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail="Device existed")

    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create a device",
        )

    if new_device.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Can't create device in group {new_device.group_id} because you are not a member of this group",
        )

    session.add(new_device)
    await session.commit()
    await session.refresh(new_device)

    return new_device


@device_router.get(path="", description="Get devices", status_code=HTTP_200_OK)
async def get_devices(
    filters: Annotated[DeviceFilters, Query()],
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[DeviceModel]:
    return await DeviceModel.filter(user, filters, session)


@device_router.delete(
    path="/{group_id}/{device_id}",
    description="Delete a device",
    status_code=HTTP_204_NO_CONTENT,
)
async def remove_device(
    group_id: str,
    device_id: str,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> None:
    if not (user.is_superuser or user.role == UserRole.ADMIN):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to delete a device",
        )

    device = await DeviceModel.get_by_key(group_id, device_id, session)

    if device.group_id not in (await user.get_group_ids(session)):
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f"Your groups do not have this device: {device_id}",
        )

    if device is not None:
        await session.delete(device)
        await session.commit()



import asyncio
import logging
import multiprocessing
import os
import signal
import sys
from functools import partial
from multiprocessing import Process

import typer
from formica import settings
from formica.executor.executor import LocalExecutor
from formica.scheduler.scheduler import Scheduler
from formica.settings import app_config
from formica.web.main import run_webserver
from sqlmodel import SQLModel

logger = logging.getLogger(__name__)
app = typer.Typer()


def handle_signal(processes: list[Process], signum, frame):
    logger.info(f"\nReceived signal {signum}, exiting gracefully...")
    for process in processes:
        if process.is_alive():
            logger.info(f"Terminating process {process.name}...")
            os.kill(process.pid, signal.SIGTERM)  # Graceful shutdown
            process.join()


@app.command()
def webserver():
    print("HELLO, STARTING WEB SERVER")
    run_webserver()


@app.command()
def scheduler():
    asyncio.run(_scheduler())


async def _scheduler():
    scheduler_ = Scheduler(LocalExecutor())
    await scheduler_.run()


async def _init():
    print("initializing database...")
    settings.init()
    async with settings.engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)
    print("database initialized")


@app.command()
def init():
    """Khởi tạo thư mục dữ liệu ở đường dẫn `FORMICA_HOME.
    Lệnh này phải được chạy trước tiên để có thể chạy được các thành phần khác"""
    asyncio.run(_init())


@app.command()
def standalone():
    init()

    # Run the processes
    web_process = multiprocessing.Process(target=webserver, daemon=False)
    scheduler_process = multiprocessing.Process(target=scheduler, daemon=True)
    print("Staring API server in as a daemon...")
    logger.info("Staring API server as a daemon...")

    web_process.start()
    scheduler_process.start()

    for sig in {signal.SIGINT, signal.SIGTERM}:
        signal.signal(sig, partial(handle_signal, [web_process, scheduler_process]))


def main():
    # multiprocessing.set_start_method("spawn")
    # logging.basicConfig()
    app()


if __name__ == "__main__":
    main()



import logging
from typing import Optional

from formica.execution.context import Context
from formica.execution.task import Task
from formica.models.constant import DeviceRunState
from formica.models.constant import TaskState
from formica.models.models import DeviceRunModel
from formica.models.models import TaskModel
from formica.node.nodes import DecisionNode

logger = logging.getLogger(__name__)


class DeviceRun:
    def __init__(self, device_run_model: DeviceRunModel, context: Context) -> None:
        self._context = context
        # TODO: Có nên hợp DB model với các class xử lý logic không? Tại vì nếu làm vậy code gọn
        #  Task và DeviceRun khi chạy có thể cập nhật trạng thái của chính mình vào DB dễ hơn
        self._device_run_model = device_run_model
        self._task_dict: dict[str, Task] = {}
        self._target_task: Optional[Task] = None
        for task_model in device_run_model.tasks:
            task = Task(
                node=context.get_node_by_id(task_model.node_id),
                trigger_rule=task_model.trigger_rule,
            )
            self._task_dict[task_model.node_id] = task

    async def run(self):
        self._context.session.add(self._device_run_model)
        self._device_run_model.set_state(DeviceRunState.RUNNING)

        while (
            next_task := await self._update_state_and_get_next_task_to_run()
        ) is not None:
            self._log_task_states()
            self._target_task = next_task
            logger.debug(f"next task is: {next_task.node.node_id}")
            target_task_model = self._find_task_model_by_node_id(next_task.node.node_id)
            target_task_model.set_state(TaskState.RUNNING)
            await self._context.session.commit()

            await next_task.run(self._context)

            # Update the state of the target task in the model
            target_task_model.set_state(next_task.state)

    def _find_entry_task(self):
        """Find task without upstream tasks, which is the entry point of the flow."""
        # TODO: Should we keep this information in Flow?
        for task in self._task_dict.values():
            upstream_tasks = [
                self._task_dict[node_id] for node_id in task.node.upstream_node_ids
            ]
            if not upstream_tasks:
                return task
        return None

    def _update_task_state(self, task: Task, state: TaskState):
        """Update the state of a task and its corresponding TaskModel."""
        task.state = state
        task_model = self._find_task_model_by_node_id(task.node.node_id)
        task_model.set_state(state)

    def _find_task_model_by_node_id(self, node_id: str) -> Optional[TaskModel]:
        for task_model in self._device_run_model.tasks:
            if task_model.node_id == node_id:
                return task_model
        return None

    def _skip_branch(self, task: Task) -> None:
        """Recursively skip the entire branch rooted at this task by setting the state of all tasks in that branch to 'SKIPPED'."""
        if not task.node.downstream_node_ids:
            self._update_task_state(task, TaskState.SKIPPED)
            return

        for downstream_task in [
            self._task_dict[node_id] for node_id in task.node.downstream_node_ids
        ]:
            self._skip_branch(downstream_task)

    async def _update_state_and_get_next_task_to_run(self) -> Optional[Task]:
        """Cập nhật trạng thái của RetryRun (đã chạy xong chưa) và tìm task tiếp theo để chạy"""
        # Nếu là toán tử vừa chạy xong là decision thì phải skip các
        # toán tử trong nhánh không thực thi và trả về toán tử sẽ được thực thi
        if self._target_task is None:
            return self._find_entry_task()

        if self._target_task.state == TaskState.FAILED:
            # Nếu Task vừa chạy mà thất bại thì không
            return None

        if isinstance(self._target_task.node, DecisionNode):
            # Tìm ra toán tử ở nhánh không được chọn
            skip_task = self._task_dict[self._target_task.node.skip_task]
            self._skip_branch(skip_task)
            await self._context.session.commit()

            return self._task_dict[self._target_task.node.output]
        else:
            if self._target_task.node.downstream_node_ids:
                return self._task_dict[self._target_task.node.downstream_node_ids[0]]

        # If control reaches here, it means all tasks are either running or finished
        # Calculate the status of this DeviceRun (success or failed)

        logging.info("Checking device run state...")
        # If only one task failed, then the whole device run is considered failed
        if any(task.state == TaskState.FAILED for task in self._task_dict.values()):
            logging.info("Set device run state to FAILED")
            self._device_run_model.set_state(DeviceRunState.FAILED)
        else:
            logging.info("Set device run state to SUCCESS")
            self._device_run_model.set_state(DeviceRunState.SUCCESS)

        await self._context.session.commit()
        return None

    def _log_task_states(self):
        """Log ra trạng thái hiện tại các task, dùng để debug"""
        logger.debug("============ TASK STATES ==============")
        for task in self._task_dict.values():
            logger.debug(f"{task.node.node_id}: {task.state}")



import asyncio
import logging
import multiprocessing
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass

from formica.models.models import DeviceRunModel
from formica.utils.workload import execute_device_run

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Workload:
    device_run_db_id: int


class Executor(ABC):
    def __init__(self):
        self._work_queue: list[Workload] = []
        self._process_list: list[multiprocessing.Process] = []

    @abstractmethod
    def run(self):
        pass

    def enqueue_device_run(self, device_run: DeviceRunModel) -> None:
        """
        Enqueue the DeviceRun to the work queue.
        :param device_run: DeviceRunModel instance to be added to the queue.
        :return: None
        :raises ValueError: RetryRun exists in the work queue
        """
        new_workload = Workload(device_run.device_run_db_id)
        if new_workload in self._work_queue:
            raise ValueError(f"DeviceRun existed: '{device_run.retry_run_db_id}'")
        self._work_queue.append(new_workload)


class LocalExecutor(Executor):
    def __init__(self):
        super().__init__()
        pass

    async def run(self, session=None):
        logger.debug(f"Executor running, found {len(self._work_queue)} workloads")
        tasks = [
            asyncio.create_task(execute_device_run(work_load.device_run_db_id))
            for work_load in self._work_queue
        ]
        await asyncio.gather(*tasks)

        # Clear the queue
        self._work_queue.clear()


class CeleryExecutor(Executor):
    def __init__(self):
        super().__init__()
        pass

    def run(self):
        pass



from typing import Annotated

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Query
from formica.models.models import FlowModel
from formica.models.models import FlowVersionModel
from formica.models.models import User
from formica.models.request_models import FlowVersionFilters
from formica.utils.session import get_session
from formica.web.users import current_active_user
from sqlmodel.ext.asyncio.session import AsyncSession
from starlette.status import HTTP_200_OK
from starlette.status import HTTP_201_CREATED
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_403_FORBIDDEN
from starlette.status import HTTP_404_NOT_FOUND

SessionDep = Annotated[AsyncSession, Depends(get_session)]
flow_version_router = APIRouter()


@flow_version_router.post(
    path="", description="Tạo một FlowVersion", status_code=HTTP_201_CREATED
)
async def add_flow_version(
    new_flow_version: FlowVersionModel,
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> FlowVersionModel:
    new_flow_version = FlowVersionModel.model_validate(new_flow_version)
    flow = await FlowModel.get_by_key(new_flow_version.flow_id, session)

    if flow is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Flow not found")

    # Only the owner or superuser can create new flow version
    if flow.owner_id != user.id and not user.is_superuser:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="You do not have permission to create flow version on this flow",
        )

    if await FlowVersionModel.flow_version_existed(
        new_flow_version.flow_id, new_flow_version.version
    ):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Flow version existed"
        )

    # valid, message = validate_flow_json(new_flow_version.structure)
    # if not valid:
    #     raise HTTPException(status_code=400, detail=f"Flow syntax error: {message}")

    new_flow_version.flow = flow
    session.add(new_flow_version)
    await session.commit()
    await session.refresh(new_flow_version)
    return new_flow_version


@flow_version_router.get(
    path="", description="Get flow versions", status_code=HTTP_200_OK
)
async def get_flow_versions(
    filters: Annotated[FlowVersionFilters, Query()],
    session: SessionDep,
    user: User = Depends(current_active_user),
) -> list[FlowVersionModel]:
    return await FlowVersionModel.filter(user, filters, session)
